from mgz.fast.header import decompress, parse_version
from mgz.summary.full import FullSummary
from mgz.model.compat import ModelSummary
from mgz.util import Version
import logging

logger = logging.getLogger(__name__)

class SummaryStub:

    def __call__(self, data, playback=None, fallback=False):
        header = decompress(data)
        version, game, save, log = parse_version(header, data)
        data.seek(0)
        if version is Version.DE and save > 13.34 and not fallback:
            logger.info("using model summary")
            try:
                return ModelSummary(data, playback)
            except RuntimeError:
                logger.warning("could not fast parse; falling back")
        logger.info("using full summary")
        return FullSummary(data, playback)


Summary = SummaryStub()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class Object(object):
    __slots__ = ['_tab']

    # Object
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Object
    def Id(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Object
    def CreatedTime(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Object
    def MasterObjectClass(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Object
    def MasterObjectId(self): return self._tab.Get(flatbuffers.number_types.Uint16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(10))
    # Object
    def State(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # Object
    def OwnerId(self): return self._tab.Get(flatbuffers.number_types.Uint8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(13))
    # Object
    def Position(self):
        from .Position import Position
        obj = Position()
        obj.Init(self._tab.Bytes, self._tab.Pos + 16)
        return obj

    # Object
    def Idle(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(28))
    # Object
    def CurrentlyCreatingMasterObjectId(self): return self._tab.Get(flatbuffers.number_types.Uint16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(30))
    # Object
    def CurrentlyResearchingTechId(self): return self._tab.Get(flatbuffers.number_types.Uint16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(32))
    # Object
    def KilledByUnitId(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(36))
    # Object
    def KilledByPlayerId(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(40))
    # Object
    def KilledTime(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(44))
    # Object
    def DeletedByOwner(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(48))
    # Object
    def BuildingPercentComplete(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(52))
    # Object
    def BuildingStartTime(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(56))
    # Object
    def BuildingCompleteTime(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(60))
    # Object
    def CumulativeIdleTime(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(64))

def CreateObject(builder, id, createdTime, masterObjectClass, masterObjectId, state, ownerId, position_x, position_y, position_z, idle, currentlyCreatingMasterObjectId, currentlyResearchingTechId, killedByUnitId, killedByPlayerId, killedTime, deletedByOwner, buildingPercentComplete, buildingStartTime, buildingCompleteTime, cumulativeIdleTime):
    builder.Prep(4, 68)
    builder.PrependFloat32(cumulativeIdleTime)
    builder.PrependUint32(buildingCompleteTime)
    builder.PrependUint32(buildingStartTime)
    builder.PrependFloat32(buildingPercentComplete)
    builder.Pad(3)
    builder.PrependBool(deletedByOwner)
    builder.PrependUint32(killedTime)
    builder.Pad(3)
    builder.PrependInt8(killedByPlayerId)
    builder.PrependInt32(killedByUnitId)
    builder.Pad(2)
    builder.PrependUint16(currentlyResearchingTechId)
    builder.PrependUint16(currentlyCreatingMasterObjectId)
    builder.Pad(1)
    builder.PrependBool(idle)
    builder.Prep(4, 12)
    builder.PrependFloat32(position_z)
    builder.PrependFloat32(position_y)
    builder.PrependFloat32(position_x)
    builder.Pad(2)
    builder.PrependUint8(ownerId)
    builder.PrependInt8(state)
    builder.PrependUint16(masterObjectId)
    builder.Pad(1)
    builder.PrependInt8(masterObjectClass)
    builder.PrependUint32(createdTime)
    builder.PrependInt32(id)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class ConfigMessage(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsConfigMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ConfigMessage()
        x.Init(buf, n + offset)
        return x

    # ConfigMessage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ConfigMessage
    def MessageIntervalMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 1000

    # ConfigMessage
    def UpdateCycles(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 1

def ConfigMessageStart(builder): builder.StartObject(2)
def ConfigMessageAddMessageIntervalMs(builder, messageIntervalMs): builder.PrependUint32Slot(0, messageIntervalMs, 1000)
def ConfigMessageAddUpdateCycles(builder, updateCycles): builder.PrependUint32Slot(1, updateCycles, 1)
def ConfigMessageEnd(builder): return builder.EndObject()



"""Determine teams."""


def get_teams_data(header):
    """Get teams."""
    allies = {}
    for number, player in enumerate(header.initial.players[1:]):
        allies[number + 1] = set([number + 1])
        for i, mode in enumerate(player.attributes.my_diplomacy):
            if mode == 'ally':
                allies[number + 1].add(i)
    return set([frozenset(s) for s in allies.values()])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class MarketCoefficients(object):
    __slots__ = ['_tab']

    # MarketCoefficients
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MarketCoefficients
    def Wood(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # MarketCoefficients
    def Food(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # MarketCoefficients
    def Stone(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateMarketCoefficients(builder, wood, food, stone):
    builder.Prep(4, 12)
    builder.PrependFloat32(stone)
    builder.PrependFloat32(food)
    builder.PrependFloat32(wood)
    return builder.Offset()



"""Fast(er) parsing for recorded game headers."""
import io
import hashlib
import re
import struct
import uuid
import zlib

from mgz.util import get_version, Version


ZLIB_WBITS = -15
CLASSES = [b'\x0a', b'\x1e', b'\x46', b'\x50']
BLOCK_END = b'\x00\x0b'
REGEXES = {}
SKIP_OBJECTS = [
    (b'\x1e\x00\x87\x02', 252) # 647: junk DE object, thousands per file
]


def _compile_object_search():
    """Compile regular expressions for object searching."""
    class_or = b'(' + b'|'.join(CLASSES) + b')'
    for i in range(9):
        expr = class_or + struct.pack('b', i) + b'[^\x00][\x00-\xff]{3}\xff\xff\xff\xff[^\xff]'
        REGEXES[i] = re.compile(expr)


_compile_object_search()


def unpack(fmt, data):
    """Unpack bytes according to format string."""
    output = struct.unpack(fmt, data.read(struct.calcsize(fmt)))
    if len(output) == 1:
        return output[0]
    return output


def aoc_string(data):
    """Read AOC string."""
    length = unpack('<h', data)
    return data.read(length)


def de_string(data):
    """Read DE string."""
    assert data.read(2) == b'\x60\x0a'
    length = unpack('<h', data)
    return unpack(f'<{length}s', data)


def hd_string(data):
    """Read HD string."""
    length = unpack('<h', data)
    assert data.read(2) == b'\x60\x0a'
    return unpack(f'<{length}s', data)


def parse_object(data, offset):
    """Parse an object."""
    class_id, object_id, instance_id, pos_x, pos_y = struct.unpack_from('<bxH14xIxff', data, offset)
    return dict(
        class_id=class_id,
        object_id=object_id,
        instance_id=instance_id,
        position=dict(
            x=pos_x,
            y=pos_y
        )
    )


def object_block(data, pos, player_number, index):
    """Parse a block of objects."""
    objects = []
    offset = None
    while True:
        if not offset:
            match = REGEXES[player_number].search(data, pos)
            end = data.find(BLOCK_END, pos) - pos + len(BLOCK_END)
            if match is None:
                break
            offset = match.start() - pos
            while end + 8 < offset:
                end += data.find(BLOCK_END, pos + end) - (pos + end) + len(BLOCK_END)
        if end + 8 == offset:
            break
        pos += offset
        # Speed optimization: Skip specified fixed-length objects.
        test = data[pos:pos + 4]
        for fingerprint, offset in SKIP_OBJECTS:
            if test == fingerprint:
                break
        else:
            objects.append(dict(parse_object(data, pos), index=index))
        offset = None
        pos += 31
    return objects, pos + end


def parse_mod(header, num_players, version):
    """Parse Userpatch mod version."""
    cur = header.tell()
    name_length = unpack(f'<xx{num_players}x36x5xh', header)
    resources = unpack(f'<{name_length + 1}xIx', header)
    values = unpack(f'<{resources}f', header)
    header.seek(cur)
    if version is Version.USERPATCH15:
        number = int(values[198])
        return number // 1000, '.'.join(list(str(number % 1000)))


def parse_player(header, player_number, num_players):
    """Parse a player (and objects)."""
    type_, *diplomacy, name_length = unpack(f'<bx{num_players}x9i5xh', header)
    name, resources = unpack(f'<{name_length - 1}s2xIx', header)
    header.read(resources * 4)
    start_x, start_y, civilization_id, color_id = unpack('<xff9xb3xbx',header)
    offset = header.tell()
    data = header.read()
    # Skips thousands of bytes that are not easy to parse.
    start = re.search(b'\x0b\x00.\x00\x00\x00\x02\x00\x00', data).end()
    objects, end = object_block(data, start, player_number, 0)
    sleeping, end = object_block(data, end, player_number, 1)
    doppel, end = object_block(data, end, player_number, 2)
    if data[end + 8:end + 10] == BLOCK_END:
        end += 10
    header.seek(offset + end)
    return dict(
        number=player_number,
        type=type_,
        name=name,
        diplomacy=diplomacy,
        civilization_id=civilization_id,
        color_id=color_id,
        objects=objects + sleeping + doppel,
        position=dict(
            x=start_x,
            y=start_y
        )
    )


def parse_lobby(data, version, save):
    """Parse lobby data."""
    if version is Version.DE:
        data.read(5)
        if save >= 20.06:
            data.read(9)
    data.read(8)
    if version not in (Version.DE, Version.HD):
        data.read(1)
    reveal_map_id, map_size, population, game_type_id, lock_teams = unpack('I4xIIbb', data)
    if version in (Version.DE, Version.HD):
        data.read(5)
        if save >= 13.13:
            data.read(4)
    chat = []
    for _ in range(0, unpack('<I', data)):
        message = data.read(unpack('<I', data)).strip(b'\x00')
        if len(message) > 0:
            chat.append(message)
    seed = None
    if version is Version.DE:
        seed = unpack('<i', data)
    return dict(
        reveal_map_id=reveal_map_id,
        map_size=map_size,
        population=population * (25 if version not in (Version.DE, Version.HD) else 1),
        game_type_id=game_type_id,
        lock_teams=lock_teams == 1,
        chat=chat,
        seed=seed
    )


def parse_map(data, version):
    """Parse map."""
    data.read(60)
    tile_format = '<xbbx'
    if version is Version.DE:
        tile_format = '<bxb6x'
        data.read(8)
    size_x, size_y, zone_num = unpack('<III', data)
    tile_num = size_x * size_y
    for _ in range(zone_num):
        if version in (Version.DE, Version.HD):
            data.read(2048 + (tile_num * 2))
        else:
            data.read(1275 + tile_num)
        num_floats = unpack('<I', data)
        data.read(num_floats * 4)
        data.read(4)
    all_visible = unpack('<bx', data)
    tiles = [unpack(tile_format, data) for _ in range(tile_num)]
    num_data = unpack('<I4x', data)
    data.read(num_data * 4)
    for _ in range(0, num_data):
        num_obs = unpack('<I', data)
        data.read(num_obs * 8)
    x2, y2 = unpack('<II', data)
    data.read(x2 * y2 * 4)
    restore_time = unpack('<I', data)
    if restore_time > 0:
        raise RuntimeError("restored matches can't be parsed yet")
    return dict(
        all_visible=all_visible == 1,
        restore_time=restore_time,
        dimension=size_x,
        tiles=tiles
    )


def parse_scenario(data, num_players, version, save):
    """Parse scenario section."""
    data.read(4455)
    if version is Version.DE:
        data.read(128)
    instructions = aoc_string(data)
    for _ in range(0, 9):
        aoc_string(data)
    data.read(78)
    for _ in range(0, 16):
        aoc_string(data)
    data.read(196)
    for _ in range(0, 16):
        data.read(24)
        if version in (Version.DE, Version.HD):
            data.read(4)
    data.read(12672)
    if version is Version.DE:
        data.read(196)
    else:
        for _ in range(0, 16):
            data.read(332)
    if version is Version.HD:
        data.read(644)
    data.read(88)
    if version is Version.HD:
        data.read(16)
    map_id, difficulty_id = unpack('<II', data)
    remainder = data.read()
    if version is Version.DE:
        if save >= 25.06:
            end = remainder.find(b'\x00\x00\x00\x00\x00\x00\x04\x40')
        else:
            end = remainder.find(b'\x33\x33\x33\x33\x33\x33\x03\x40')
        end += 1045
    else:
        end = remainder.find(b'\x9a\x99\x99\x99\x99\x99\xf9\x3f') + 13
    data.seek(end - len(remainder), 1)
    return dict(
        map_id=map_id,
        difficulty_id=difficulty_id,
        instructions=instructions
    )


def parse_de(data, version, save, skip=False):
    """Parse DE-specific header."""
    if version is not Version.DE:
        return None
    data.read(12)
    dlc_count = unpack('<I', data)
    data.read(dlc_count * 4)
    data.read(4)
    difficulty_id = unpack('<I', data)
    data.read(20)
    starting_age_id = unpack('<I', data)
    data.read(46)
    random_positions, all_technologies = unpack('<bb', data)
    data.read(2)
    lock_speed = unpack('<b', data)
    data.read(20)
    if save >= 25.06:
        data.read(1)
    players = []
    for _ in range(8):
        data.read(4)
        color_id = unpack('<i', data)
        data.read(12)
        civilization_id = unpack('<I', data)
        de_string(data)
        data.read(1)
        de_string(data)
        name = de_string(data)
        data.read(4)
        profile_id, number = unpack('<I4xi', data)
        data.read(8)
        prefer_random = unpack('b', data)
        data.read(1)
        if save >= 25.06:
            data.read(8)
        if name:
            players.append(dict(
                number=number,
                color_id=color_id,
                name=name,
                profile_id=profile_id,
                civilization_id=civilization_id,
                prefer_random=prefer_random == 1
            ))
    data.read(33)
    for _ in range(23):
        de_string(data)
        c = unpack('<I', data)
        while c in [3, 21, 23, 42, 44, 45, 46, 47]:
            c = unpack('<I', data)
    data.read(236)
    for _ in range(unpack('<Q', data)):
        data.read(4)
        de_string(data)
        data.read(4)
    if save >= 25.02:
        data.read(8)
    guid = data.read(16)
    lobby = de_string(data)
    mod = de_string(data)
    data.read(33)
    if save >= 20.06:
        data.read(1)
    if save >= 20.16:
        data.read(8)
    if save >= 25.06:
        data.read(21)
    if not skip:
        de_string(data)
        data.read(8)
    return dict(
        players=players,
        guid=str(uuid.UUID(bytes=guid)),
        hash=hashlib.sha1(guid),
        lobby=lobby.decode('utf-8'),
        mod=mod.decode('utf-8'),
        difficulty_id=difficulty_id,
        starting_age_id=starting_age_id - 2 if starting_age_id > 0 else 0,
        team_together=not bool(random_positions),
        all_technologies=bool(all_technologies),
        lock_speed=bool(lock_speed)
    )


def parse_hd(data, version, save):
    """Parse HD-specifc header."""
    if version is not Version.HD or save <= 12.34:
        return None
    data.read(12)
    dlc_count = unpack('<I', data)
    data.read(dlc_count * 4)
    data.read(4)
    difficulty_id, map_id = unpack('<II', data)
    data.read(80)
    players = []
    for _ in range(8):
        data.read(4)
        color_id = unpack('<i', data)
        data.read(12)
        civilization_id = unpack('<I', data)
        hd_string(data)
        data.read(1)
        hd_string(data)
        name = hd_string(data)
        data.read(4)
        steam_id, number = unpack('<Qi', data)
        data.read(8)
        if name:
            players.append(dict(
                number=number,
                color_id=color_id,
                name=name,
                profile_id=steam_id,
                civilization_id=civilization_id
            ))
    data.read(26)
    hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(8)
    guid = data.read(16)
    lobby = hd_string(data)
    mod = hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(4)
    return dict(
        players=players,
        guid=str(uuid.UUID(bytes=guid)),
        lobby=lobby.decode('utf-8'),
        mod=mod.decode('utf-8'),
        map_id=map_id,
        difficulty_id=difficulty_id
    )


def decompress(data):
    """Decompress header bytes."""
    prefix_size = 8
    header_len, _ = unpack('<II', data)
    zlib_header = data.read(header_len - prefix_size)
    return io.BytesIO(zlib.decompress(zlib_header, wbits=ZLIB_WBITS))


def parse_version(header, data):
    """Parse and compute game version."""
    log = unpack('<I', data)
    game, save = unpack('<7sxf', header)
    version = get_version(game.decode('ascii'), round(save, 2), log)
    return version, game.decode('ascii'), round(save, 2), log


def parse_players(header, num_players, version):
    """Parse all players."""
    cur = header.tell()
    gaia = b'Gaia' if version in (Version.DE, Version.HD) else b'GAIA'
    anchor = header.read().find(b'\x05\x00' + gaia + b'\x00')
    header.seek(cur + anchor - num_players - 43)
    mod = parse_mod(header, num_players, version)
    players = [parse_player(header, number, num_players) for number in range(num_players)]
    for _ in range(num_players):
        header.read(4)
        entries = unpack('<i', header)
        header.read(5 + (entries * 44))
        points = unpack('<i', header)
        header.read(8 + (points * 32))
    return players, mod


def parse_metadata(header):
    """Parse recorded game metadata."""
    ai, game_speed, owner_id, num_players, cheats = unpack('<I24xf17xhbxb', header)
    if ai > 0:
        raise RuntimeError("don't know how to parse ai")
    return dict(
        speed=game_speed,
        owner_id=owner_id,
        cheats=cheats == 1
    ), num_players


def parse(data):
    """Parse recorded game header."""
    try:
        header = decompress(data)
        version, game, save, log = parse_version(header, data)
        if version not in (Version.USERPATCH15, Version.DE, Version.HD):
            raise RuntimeError(f"{version} not supported")
        de = parse_de(header, version, save)
        hd = parse_hd(header, version, save)
        metadata, num_players = parse_metadata(header)
        map_ = parse_map(header, version)
        players, mod = parse_players(header, num_players, version)
        scenario = parse_scenario(header, num_players, version, save)
        lobby = parse_lobby(header, version, save)
    except (struct.error, zlib.error, AssertionError, MemoryError):
        raise RuntimeError("could not parse")
    return dict(
        version=version,
        game_version=game,
        save_version=save,
        log_version=log,
        players=players,
        map=map_,
        de=de,
        hd=hd,
        mod=mod,
        metadata=metadata,
        scenario=scenario,
        lobby=lobby
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class Player(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Player()
        x.Init(buf, n + offset)
        return x

    # Player
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Player
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Player
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Player
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Player
    def Food(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def Wood(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def Stone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def Gold(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def Headroom(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def Population(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def PopulationCap(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def CumulativeHousedTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def CumulativePopCappedTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def CivilianPopulation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def MilitaryPopulation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def PercentMapExplored(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Player
    def VictoryPointsAndAttributes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            x = o + self._tab.Pos
            from .VictoryPointsAndAttributes import VictoryPointsAndAttributes
            obj = VictoryPointsAndAttributes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Player
    def Techs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 12
            from .Tech import Tech
            obj = Tech()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Player
    def TechsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def PlayerStart(builder): builder.StartObject(17)
def PlayerAddId(builder, id): builder.PrependUint8Slot(0, id, 0)
def PlayerAddName(builder, name): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def PlayerAddState(builder, state): builder.PrependInt8Slot(2, state, 0)
def PlayerAddFood(builder, food): builder.PrependFloat32Slot(3, food, 0.0)
def PlayerAddWood(builder, wood): builder.PrependFloat32Slot(4, wood, 0.0)
def PlayerAddStone(builder, stone): builder.PrependFloat32Slot(5, stone, 0.0)
def PlayerAddGold(builder, gold): builder.PrependFloat32Slot(6, gold, 0.0)
def PlayerAddHeadroom(builder, headroom): builder.PrependFloat32Slot(7, headroom, 0.0)
def PlayerAddPopulation(builder, population): builder.PrependFloat32Slot(8, population, 0.0)
def PlayerAddPopulationCap(builder, populationCap): builder.PrependFloat32Slot(9, populationCap, 0.0)
def PlayerAddCumulativeHousedTime(builder, cumulativeHousedTime): builder.PrependFloat32Slot(10, cumulativeHousedTime, 0.0)
def PlayerAddCumulativePopCappedTime(builder, cumulativePopCappedTime): builder.PrependFloat32Slot(11, cumulativePopCappedTime, 0.0)
def PlayerAddCivilianPopulation(builder, civilianPopulation): builder.PrependFloat32Slot(12, civilianPopulation, 0.0)
def PlayerAddMilitaryPopulation(builder, militaryPopulation): builder.PrependFloat32Slot(13, militaryPopulation, 0.0)
def PlayerAddPercentMapExplored(builder, percentMapExplored): builder.PrependFloat32Slot(14, percentMapExplored, 0.0)
def PlayerAddVictoryPointsAndAttributes(builder, victoryPointsAndAttributes): builder.PrependStructSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(victoryPointsAndAttributes), 0)
def PlayerAddTechs(builder, techs): builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(techs), 0)
def PlayerStartTechsVector(builder, numElems): return builder.StartVector(12, numElems, 4)
def PlayerEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class Tech(object):
    __slots__ = ['_tab']

    # Tech
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Tech
    def IdIndex(self): return self._tab.Get(flatbuffers.number_types.Int16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Tech
    def IdGame(self): return self._tab.Get(flatbuffers.number_types.Int16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # Tech
    def Time(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Tech
    def State(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateTech(builder, idIndex, idGame, time, state):
    builder.Prep(4, 12)
    builder.Pad(3)
    builder.PrependInt8(state)
    builder.PrependUint32(time)
    builder.PrependInt16(idGame)
    builder.PrependInt16(idIndex)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

class TechState(object):
    CantResearch = 0
    CanResearchLater = 1
    CanResearch = 2
    Researching = 3
    Researched = 4




"""Definitive Edition structure."""

from construct import (
    Struct, Int32ul, Float32l, Array, Padding, Flag, If,
    Byte, Int16ul, Bytes, Int32sl, Peek, Const, RepeatUntil,
    Int64ul, Computed
)

from mgz.enums import VictoryEnum, ResourceLevelEnum, AgeEnum, PlayerTypeEnum, DifficultyEnum
from mgz.util import find_save_version

# pylint: disable=invalid-name, bad-continuation

de_string = Struct(
    Const(b"\x60\x0A"),
    "length"/Int16ul,
    "value"/Bytes(lambda ctx: ctx.length)
)

separator = Const(b"\xa3_\x02\x00")

player = Struct(
    "dlc_id"/Int32ul,
    "color_id"/Int32sl,
    "selected_color"/Byte,
    "selected_team_id"/Byte,
    "resolved_team_id"/Byte,
    "dat_crc"/Bytes(8),
    "mp_game_version"/Byte,
    "civ_id"/Int32ul,
    "ai_type"/de_string,
    "ai_civ_name_index"/Byte,
    "ai_name"/de_string,
    "name"/de_string,
    "type"/PlayerTypeEnum(Int32ul),
    "profile_id"/Int32ul,
    Const(b"\x00\x00\x00\x00"),
    "player_number"/Int32sl,
    "hd_rm_elo"/If(lambda ctx: find_save_version(ctx) < 25.22, Int32ul),
    "hd_dm_elo"/If(lambda ctx: find_save_version(ctx) < 25.22, Int32ul),
    "prefer_random"/Flag,
    "custom_ai"/Flag,
    If(lambda ctx: find_save_version(ctx) >= 25.06, "handicap"/Bytes(8)),
)

de = "de"/Struct(
    "build"/If(lambda ctx: find_save_version(ctx) >= 25.22, Int32ul),
    "version"/Float32l,
    "interval_version"/Int32ul,
    "game_options_version"/Int32ul,
    "dlc_count"/Int32ul,
    "dlc_ids"/Array(lambda ctx: ctx.dlc_count, Int32ul),
    "dataset_ref"/Int32ul,
    Peek("difficulty_id"/Int32ul),
    DifficultyEnum("difficulty"/Int32ul),
    "selected_map_id"/Int32ul,
    "resolved_map_id"/Int32ul,
    "reveal_map"/Int32ul,
    Peek("victory_type_id"/Int32ul),
    VictoryEnum("victory_type"/Int32ul),
    Peek("starting_resources_id"/Int32ul),
    ResourceLevelEnum("starting_resources"/Int32ul),
    "starting_age_id"/Int32ul,
    "starting_age"/AgeEnum(Computed(lambda ctx: ctx.starting_age_id - 2)),
    "ending_age_id"/Int32ul,
    "ending_age"/AgeEnum(Computed(lambda ctx: ctx.ending_age_id - 2)),
    "game_type"/Int32ul,
    separator,
    separator,
    "speed"/Float32l,
    "treaty_length"/Int32ul,
    "population_limit"/Int32ul,
    "num_players"/Int32ul,
    "unused_player_color"/Int32ul,
    "victory_amount"/Int32ul,
    separator,
    "trade_enabled"/Flag,
    "team_bonus_disabled"/Flag,
    "random_positions"/Flag,
    "all_techs"/Flag,
    "num_starting_units"/Byte,
    "lock_teams"/Flag,
    "lock_speed"/Flag,
    "multiplayer"/Flag,
    "cheats"/Flag,
    "record_game"/Flag,
    "animals_enabled"/Flag,
    "predators_enabled"/Flag,
    "turbo_enabled"/Flag,
    "shared_exploration"/Flag,
    "team_positions"/Flag,
    "sub_game_mode"/If(lambda ctx: find_save_version(ctx) >= 13.34, Int32ul),
    "battle_royale_time"/If(lambda ctx: find_save_version(ctx) >= 13.34, Int32ul),
    "handicap"/If(lambda ctx: find_save_version(ctx) >= 25.06, Flag),
    separator,
    "players"/Array(8, player),
    Bytes(9),
    "fog_of_war"/Flag,
    "cheat_notifications"/Flag,
    "colored_chat"/Flag,
    separator,
    "ranked"/Flag,
    "allow_specs"/Flag,
    "lobby_visibility"/Int32ul,
    "hidden_civs"/Flag,
    "matchmaking"/Flag,
    "spec_delay"/Int32ul,
    "scenario_civ"/If(lambda ctx: ctx._.save_version >= 13.13, Byte),
    "rms_crc"/If(lambda ctx: ctx._.save_version >= 13.13, Bytes(4)),
    "strings"/Array(23,
        Struct(
            "string"/de_string,
            RepeatUntil(lambda x, lst, ctx: lst[-1] not in [3, 21, 23, 42, 44, 45, 46, 47], Int32ul)
        )
    ),
    "strategic_numbers"/If(lambda ctx: ctx._.save_version < 25.22, Array(59, Int32sl)),
    "num_ai_files"/Int64ul,
    "ai_files"/Array(lambda ctx: ctx.num_ai_files, Struct(
        Bytes(4),
        "name"/de_string,
        Bytes(4),
    )),
    If(lambda ctx: ctx._.save_version >= 25.02, Bytes(8)),
    "guid"/Bytes(16),
    "lobby_name"/de_string,
    If(lambda ctx: ctx._.save_version >= 25.22, Bytes(8)),
    "modded_dataset"/de_string,
    Bytes(19),
    If(lambda ctx: ctx._.save_version >= 13.13, Bytes(5)),
    If(lambda ctx: ctx._.save_version >= 13.17, Bytes(9)),
    If(lambda ctx: ctx._.save_version >= 20.06, Bytes(1)),
    If(lambda ctx: ctx._.save_version >= 20.16, Bytes(8)),
    If(lambda ctx: ctx._.save_version >= 25.06, Bytes(21)),
    If(lambda ctx: ctx._.save_version >= 25.22, Bytes(4)),
    de_string,
    Bytes(5),
    If(lambda ctx: ctx._.save_version >= 13.13, Byte),
    If(lambda ctx: ctx._.save_version < 13.17, Struct(
        de_string,
        Int32ul,
        Bytes(4), # usually zero
    )),
    If(lambda ctx: ctx._.save_version >= 13.17, Bytes(2))
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class VictoryPointsAndAttributes(object):
    __slots__ = ['_tab']

    # VictoryPointsAndAttributes
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # VictoryPointsAndAttributes
    def VpRelicsCaptured(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # VictoryPointsAndAttributes
    def VpRelicGold(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # VictoryPointsAndAttributes
    def VpTradeProfit(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # VictoryPointsAndAttributes
    def VpTributeReceived(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # VictoryPointsAndAttributes
    def VpTributeSent(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # VictoryPointsAndAttributes
    def VpTotalFood(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(20))
    # VictoryPointsAndAttributes
    def VpTotalWood(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(24))
    # VictoryPointsAndAttributes
    def VpTotalGold(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(28))
    # VictoryPointsAndAttributes
    def VpTotalStone(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(32))
    # VictoryPointsAndAttributes
    def AttrValueSpentObjects(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(36))
    # VictoryPointsAndAttributes
    def AttrValueSpentResearch(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(40))
    # VictoryPointsAndAttributes
    def AttrValueLostUnits(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(44))
    # VictoryPointsAndAttributes
    def AttrValueLostBuildings(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(48))
    # VictoryPointsAndAttributes
    def AttrValueCurrentUnits(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(52))
    # VictoryPointsAndAttributes
    def AttrValueCurrentBuildings(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(56))
    # VictoryPointsAndAttributes
    def AttrValueObjectsDestroyed(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(60))
    # VictoryPointsAndAttributes
    def AttrKills(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(64))
    # VictoryPointsAndAttributes
    def AttrDeaths(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(68))
    # VictoryPointsAndAttributes
    def AttrRazes(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(72))

def CreateVictoryPointsAndAttributes(builder, vpRelicsCaptured, vpRelicGold, vpTradeProfit, vpTributeReceived, vpTributeSent, vpTotalFood, vpTotalWood, vpTotalGold, vpTotalStone, attrValueSpentObjects, attrValueSpentResearch, attrValueLostUnits, attrValueLostBuildings, attrValueCurrentUnits, attrValueCurrentBuildings, attrValueObjectsDestroyed, attrKills, attrDeaths, attrRazes):
    builder.Prep(4, 76)
    builder.PrependFloat32(attrRazes)
    builder.PrependFloat32(attrDeaths)
    builder.PrependFloat32(attrKills)
    builder.PrependFloat32(attrValueObjectsDestroyed)
    builder.PrependFloat32(attrValueCurrentBuildings)
    builder.PrependFloat32(attrValueCurrentUnits)
    builder.PrependFloat32(attrValueLostBuildings)
    builder.PrependFloat32(attrValueLostUnits)
    builder.PrependFloat32(attrValueSpentResearch)
    builder.PrependFloat32(attrValueSpentObjects)
    builder.PrependFloat32(vpTotalStone)
    builder.PrependFloat32(vpTotalGold)
    builder.PrependFloat32(vpTotalWood)
    builder.PrependFloat32(vpTotalFood)
    builder.PrependFloat32(vpTributeSent)
    builder.PrependFloat32(vpTributeReceived)
    builder.PrependFloat32(vpTradeProfit)
    builder.PrependFloat32(vpRelicGold)
    builder.PrependFloat32(vpRelicsCaptured)
    return builder.Offset()



"""Fast(er) parsing for situations requiring speed."""
import struct
from enum import Enum

from mgz.util import check_flags

CHECKSUM_INTERVAL = 500

class Operation(Enum):
    """Operation types."""
    ACTION = 1
    SYNC = 2
    VIEWLOCK = 3
    CHAT = 4
    START = 5
    SAVE = 6

class Action(Enum):
    """Action types."""
    ORDER = 0
    STOP = 1
    WORK = 2
    MOVE = 3
    CREATE = 4
    ADD_ATTRIBUTE = 5
    GIVE_ATTRIBUTE = 6
    AI_ORDER = 10
    RESIGN = 11
    SPECTATE = 15
    ADD_WAYPOINT = 16
    STANCE = 18
    GUARD = 19
    FOLLOW = 20
    PATROL = 21
    FORMATION = 23
    SAVE = 27
    GROUP_MULTI_WAYPOINTS = 31
    CHAPTER = 32
    DE_ATTACK_MOVE = 33
    HD_UNKNOWN_34 = 34
    DE_UNKNOWN_35 = 35
    DE_UNKNOWN_37 = 37
    DE_AUTOSCOUT = 38
    DE_UNKNOWN_39 = 39
    DE_UNKNOWN_41 = 41
    AI_COMMAND = 53
    MAKE = 100
    RESEARCH = 101
    BUILD = 102
    GAME = 103
    WALL = 105
    DELETE = 106
    ATTACK_GROUND = 107
    TRIBUTE = 108
    DE_UNKNOWN_109 = 109
    REPAIR = 110
    UNGARRISON = 111
    MULTIQUEUE = 112
    GATE = 114
    FLARE = 115
    SPECIAL = 117
    QUEUE = 119
    GATHER_POINT = 120
    SELL = 122
    BUY = 123
    DROP_RELIC = 126
    TOWN_BELL = 127
    BACK_TO_WORK = 128
    DE_QUEUE = 129
    DE_UNKNOWN_130 = 130
    DE_UNKNOWN_131 = 131
    DE_UNKNOWN_135 = 135
    DE_UNKNOWN_196 = 196
    POSTGAME = 255


def sync(data):
    """Handle synchronizations."""
    increment, marker = struct.unpack('<II', data.read(8))
    data.seek(-4, 1)
    checksum = None
    if marker == 0:
        data.read(8)
        checksum, = struct.unpack('<I', data.read(4))
        data.read(4)
        seq, = struct.unpack('<I', data.read(4))
        if seq > 0:
            data.read(332)
        data.read(8)
    return increment, checksum


def viewlock(data):
    """Handle viewlocks."""
    x, y, _ = struct.unpack('<ffI', data.read(12))
    return x, y


def parse_action(action_type, data):
    """Parse player, objects, and coordinates from actions."""
    if action_type == Action.RESIGN:
        return dict(player_id=data[0])
    if action_type == Action.TRIBUTE:
        player_id, player_id_to, resource_id, amount, fee = struct.unpack_from('<bbbff', data)
        return dict(player_id=player_id, player_id_to=player_id_to, resource_id=resource_id, amount=amount, fee=fee)
    if action_type == Action.MOVE:
        player_id, selected, x, y = struct.unpack_from('<b6xI2f', data)
        object_ids = []
        if selected != 255:
            offset = 0
            if check_flags(struct.unpack_from('<4b', data[19:])):
                offset = 4
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[19 + offset:])
        return dict(player_id=player_id, x=x, y=y, object_ids=list(object_ids))
    if action_type == Action.CREATE:
        player_id, x, y = struct.unpack_from('<3xhx2f', data)
        return dict(player_id=player_id, x=x, y=y)
    if action_type == Action.ORDER:
        player_id, target_id, selected, x, y = struct.unpack_from('<b2xIh2x2f', data)
        object_ids = []
        if selected != 255:
            offset = 0
            if check_flags(struct.unpack_from('<4b', data[19:])):
                offset = 4
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[19 + offset:])
        return dict(player_id=player_id, target_id=target_id, x=x, y=y, object_ids=list(object_ids))
    if action_type == Action.BUILD:
        player_id, x, y, building_id = struct.unpack_from('<xh2fI', data)
        return dict(player_id=player_id, x=x, y=y, building_id=building_id)
    if action_type == Action.STANCE:
        stance_id, *object_ids = struct.unpack_from('<xb' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), stance_id=stance_id)
    if action_type == Action.RESEARCH:
        object_id, player_id = struct.unpack_from('<3xIh', data)
        if len(data) >= 19:
            technology_id, = struct.unpack_from('<I', data[11:15])
        else:
            technology_id, = struct.unpack_from('<h', data[9:11])
        return dict(player_id=player_id, technology_id=technology_id, object_ids=[object_id])
    if action_type == Action.FORMATION:
        player_id, formation_id, *object_ids = struct.unpack_from('<xhI' + str(data[0]) + 'I', data)
        return dict(player_id=player_id, object_ids=list(object_ids), formation_id=formation_id)
    if action_type == Action.QUEUE:
        object_id, unit_id, amount = struct.unpack_from('<3xIhh', data)
        return dict(object_ids=[object_id], unit_id=unit_id, amount=amount)
    if action_type == Action.GATHER_POINT:
        target_id, x, y, *object_ids = struct.unpack_from('<3xi4x2f' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y, target_id=target_id)
    if action_type == Action.MULTIQUEUE:
        unit_id, amount, *object_ids = struct.unpack_from('<3xhxb' + str(data[5]) + 'I', data)
        return dict(object_ids=object_ids, unit_id=unit_id, amount=amount)
    if action_type == Action.PATROL:
        x, y, *object_ids = struct.unpack_from('<3xf36xf36x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.SPECIAL:
        target_id, order_id, x, y, *flags = struct.unpack_from('<3xib3x2f4x4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        object_ids = struct.unpack_from('<' + str(data[0]) + 'I', data[23 + offset:])
        values = dict(object_ids=list(object_ids), order_id=order_id)
        if x > 0 and y > 0:
            values.update(dict(x=x, y=y))
        if target_id > 0:
            values.update(dict(target_id=target_id))
        return values
    if action_type == Action.BACK_TO_WORK:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.UNGARRISON:
        selected, = struct.unpack_from('<h', data)
        x, y, *object_ids = struct.unpack_from('<3x2f8x' + str(selected) + 'I', data)
        if x > 0 and y > 0:
            return dict(object_ids=list(object_ids), x=x, y=y)
        return dict(object_ids=list(object_ids))
    if action_type == Action.BUY:
        player_id, resource_id, amount = struct.unpack_from('<bbb', data)
        return dict(player_id=player_id, resource_id=resource_id, amount=amount)
    if action_type == Action.SELL:
        player_id, resource_id, amount = struct.unpack_from('<bbb', data)
        return dict(player_id=player_id, resource_id=resource_id, amount=amount)
    if action_type == Action.DELETE:
        object_id, player_id = struct.unpack_from('<3x2I', data)
        return dict(player_id=player_id, object_ids=[object_id])
    if action_type == Action.TOWN_BELL:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.WALL:
        player_id, x, y = struct.unpack_from('<x3b', data)
        return dict(player_id=player_id, x=x, y=y)
    if action_type == Action.GAME:
        return dict(player_id=data[1], command_id=data[0])
    if action_type == Action.FLARE:
        x, y, player_id = struct.unpack_from('<19x2fb', data)
        return dict(player_id=player_id, x=x, y=y)
    if action_type == Action.REPAIR:
        target_id, *flags = struct.unpack_from('<3xI4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        object_ids = struct.unpack_from('<' + str(data[0]) + 'I', data[7 + offset:])
        return dict(target_id=target_id, object_ids=list(object_ids))
    if action_type == Action.STOP:
        object_ids = struct.unpack_from('<x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    if action_type == Action.GATE:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.FOLLOW:
        object_ids = struct.unpack_from('<7x' + str(data[0]) + 'I', data)
        return dict(object_ids=object_ids)
    if action_type == Action.GUARD:
        object_ids = struct.unpack('<7x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    if action_type == Action.ATTACK_GROUND:
        object_ids = []
        selected, x, y, *flags = struct.unpack_from('<b2x2f4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        if selected > 0:
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[11 + offset:])
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.ADD_WAYPOINT:
        object_ids = []
        selected, x, y = struct.unpack_from('<xb2b', data)
        if selected > 0:
            object_ids = struct.unpack_from('<4x' + str(selected) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.DE_QUEUE:
        player_id, unit_id, amount, *object_ids = struct.unpack_from('<b4xhbx' + str(data[3]) + 'I', data)
        return dict(player_id=player_id, object_ids=object_ids, amount=amount, unit_id=unit_id)
    if action_type == Action.DE_ATTACK_MOVE:
        x, y, *object_ids = struct.unpack_from('<3xf36xf36x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.DE_AUTOSCOUT:
        object_ids = struct.unpack_from('<x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    return dict()


def action(data):
    """Handle actions."""
    length, = struct.unpack('<I', data.read(4))
    action_id = int.from_bytes(data.read(1), 'little')
    action_bytes = data.read(length - 1)
    data.read(4)
    action_type = Action(action_id)
    if action_type == Action.POSTGAME:
        payload = dict(bytes=action_bytes + data.read())
    else:
        payload = parse_action(action_type, action_bytes)
    return action_type, payload


def chat(data):
    """Handle chat."""
    _, length = struct.unpack('<II', data.read(8))
    msg = data.read(length)
    return msg


def start(data):
    """Handle start."""
    data.read(20)
    a, b, _ = struct.unpack('<III', data.read(12))
    if a != 0: # AOC 1.0x
        data.seek(-12, 1)
    if b == 2: # DE
        data.seek(-8, 1)


def save(data):
    """Handle saved chapter."""
    data.seek(-4, 1)
    pos = data.tell()
    length, _ = struct.unpack('<II', data.read(8))
    data.read(length - pos - 8)


def meta(data):
    """Handle log meta."""
    try:
        first, = struct.unpack('<I', data.read(4))
        if first != 500: # Not AOK
            data.read(4)
        data.read(20)
        a, b, _ = struct.unpack('<III', data.read(12))
        if a != 0: # AOC 1.0x
            data.seek(-12, 1)
        if b == 2: # DE
            data.seek(-8, 1)
    except struct.error:
        raise ValueError("insufficient meta received")


def operation(data):
    """Handle body operations."""
    try:
        op_id, = struct.unpack('<I', data.read(4))
        try:
            op_type = Operation(op_id)
        except ValueError:
            return Operation.SAVE, save(data)
        if op_type == Operation.ACTION:
            return op_type, action(data)
        if op_type == Operation.SYNC:
            return op_type, sync(data)
        if op_type == Operation.VIEWLOCK:
            return op_type, viewlock(data)
        if op_type == Operation.CHAT:
            return op_type, chat(data)
    except struct.error:
        raise EOFError
    raise RuntimeError("unknown data received")



""""Extract data via playback."""

import math
import logging
import time
from collections import defaultdict, deque
from datetime import timedelta

from mgz import fast
from mgz.playback import Client, Source


LOGGER = logging.getLogger(__name__)
CLASS_UNIT = 70
CLASS_BUILDING = 80
TECH_STATE_AVAILABLE = 2
TECH_STATE_RESEARCHING = 3
TECH_STATE_DONE = 4
DE_TECH_STATE_AVAILABLE = 1
DE_TECH_STATE_RESEARCHING = 2
DE_TECH_STATE_DONE = 3
DE_TECH_STATE_CANT_RESEARCH = -1
STATUS_WINNER = 1


def has_diff(data, **kwargs):
    """Simple dict subset diff."""
    for key, value in kwargs.items():
        if data.get(key) != value:
            return True
    return False


def flatten_research(research):
    """Flatten research data."""
    flat = []
    for pid, techs in research.items():
        for tid, values in techs.items():
            flat.append(dict(values, player_number=pid, technology_id=tid))
    return flat


def build_fb_timeseries_record(tick, player):
    """Build a timeseries record."""
    return {
        'timestamp': tick,
        'player_number': player.Id(),
        'population': player.Population(),
        'military': player.MilitaryPopulation(),
        'percent_explored': player.PercentMapExplored(),
        'headroom': int(player.Headroom()),
        'food': int(player.Food()),
        'wood': int(player.Wood()),
        'stone': int(player.Stone()),
        'gold': int(player.Gold()),
        'total_housed_time': int(player.CumulativeHousedTime()),
        'total_popcapped_time': int(player.CumulativePopCappedTime()),
        'relics_captured': int(player.VictoryPointsAndAttributes().VpRelicsCaptured()),
        'relic_gold': int(player.VictoryPointsAndAttributes().VpRelicGold()),
        'trade_profit': int(player.VictoryPointsAndAttributes().VpTradeProfit()),
        'tribute_received': int(player.VictoryPointsAndAttributes().VpTributeReceived()),
        'tribute_sent': int(player.VictoryPointsAndAttributes().VpTributeSent()),
        'total_food': int(player.VictoryPointsAndAttributes().VpTotalFood()),
        'total_wood': int(player.VictoryPointsAndAttributes().VpTotalWood()),
        'total_gold': int(player.VictoryPointsAndAttributes().VpTotalGold()),
        'total_stone': int(player.VictoryPointsAndAttributes().VpTotalStone()),
        'value_spent_objects': int(player.VictoryPointsAndAttributes().AttrValueSpentObjects()),
        'value_spent_research': int(player.VictoryPointsAndAttributes().AttrValueSpentResearch()),
        'value_lost_units': int(player.VictoryPointsAndAttributes().AttrValueLostUnits()),
        'value_lost_buildings': int(player.VictoryPointsAndAttributes().AttrValueLostBuildings()),
        'value_current_units': int(player.VictoryPointsAndAttributes().AttrValueCurrentUnits()),
        'value_current_buildings': int(player.VictoryPointsAndAttributes().AttrValueCurrentBuildings()),
        'value_objects_destroyed': int(player.VictoryPointsAndAttributes().AttrValueObjectsDestroyed()),
        'kills': int(player.VictoryPointsAndAttributes().AttrKills()),
        'deaths': int(player.VictoryPointsAndAttributes().AttrDeaths()),
        'razes': int(player.VictoryPointsAndAttributes().AttrRazes())
    }


def build_json_timeseries_record(tick, player):
    """Build a timeseries record."""
    return {
        'timestamp': tick,
        'player_number': player['id'],
        'population': player['attributes']['population'],
        'military': player['attributes']['militaryPopulation'],
        'percent_explored': player['attributes']['percentMapExplored'],
        'headroom': int(player['attributes']['populationFreeRoom']),
        'food': int(player['attributes']['food']),
        'wood': int(player['attributes']['wood']),
        'stone': int(player['attributes']['stone']),
        'gold': int(player['attributes']['gold']),
        'total_housed_time': int(player['cumulativeHousedTime']),
        'total_popcapped_time': int(player['cumulativePopCappedTime']),
        'relics_captured': int(player['attributes']['relicsCaptured']),
        'relic_gold': int(player['attributes']['relicIncomeSum']),
        'trade_profit': int(player['attributes']['tradeIncomeSum']),
        'tribute_received': int(player['attributes']['tributeReceived']),
        'tribute_sent': int(player['attributes']['tributeSent']),
        'total_food': int(player['attributes']['foodTotalGathered']),
        'total_wood': int(player['attributes']['woodTotalGathered']),
        'total_gold': int(player['attributes']['goldTotalGathered']),
        'total_stone': int(player['attributes']['stoneTotalGathered']),
        'value_spent_objects': int(player['attributes']['objectCostSum']),
        'value_spent_research': int(player['attributes']['techCostSum']),
        'value_lost_units': int(player['attributes']['unitsLostValue']),
        'value_lost_buildings': int(player['attributes']['buildingsLostValue']),
        'value_current_units': int(player['attributes']['valueOfArmy']),
        'value_current_buildings': int(player['attributes']['valueOfBuildings']),
        'value_objects_destroyed': int(player['attributes']['killsValue'] + player['attributes']['razingsValue']),
        'value_units_killed': int(player['attributes']['killsValue']),
        'value_buildings_razed': int(player['attributes']['razingsValue']),
        'trained': int(player['attributes']['totalUnitsTrained']),
        'converted': int(player['attributes']['unitsConverted']),
        'kills': int(player['attributes']['unitsKilled']),
        'deaths': int(player['attributes']['unitsLost']),
        'razes': int(player['attributes']['razings']),
        'buildings_lost': int(player['attributes']['buildingsLost']),
        'hit_points_razed': int(player['attributes']['hitPointsRazed']),
        'hit_points_killed': int(player['attributes']['hitPointsKilled']),
        'villager_high': int(player['victoryPoints']['maxVillagers']),
        'military_high': int(player['victoryPoints']['maxMilitary']),
        'total_score': int(player['victoryPoints']['total']),
        'military_score': int(player['victoryPoints']['military']),
        'economy_score': int(player['victoryPoints']['economy']),
        'society_score': int(player['victoryPoints']['society']),
        'technology_score': int(player['victoryPoints']['technology'])
    }


def update_research(player_number, state, index, tick, research, researching_state, done_state, available_state):
    """Update research structure."""
    if state == researching_state and index not in research[player_number]:
        research[player_number][index] = {
            'started': tick,
            'finished': None
        }
    elif state == done_state and index in research[player_number]:
        research[player_number][index]['finished'] = tick
    elif state == available_state and index in research[player_number]:
        del research[player_number][index]


def update_market(tick, food, wood, stone, market):
    """Update market coefficients structure."""
    last = None
    change = True
    if market:
        last = market[-1]
        change = has_diff(last, food=food, wood=wood, stone=stone)
    if change:
        market.append({
            'timestamp': tick,
            'food': food,
            'wood': wood,
            'stone': stone
        })


def add_objects(owner_id, instance_id, class_id, object_id, created, pos_x, pos_y, objects):
    """Add objects."""
    player_number = owner_id if owner_id > 0 else None
    if class_id not in [CLASS_UNIT, CLASS_BUILDING]:
        return
    if instance_id not in objects:
        objects[instance_id] = {
            'initial_player_number': player_number,
            'initial_object_id': object_id,
            'initial_class_id': class_id,
            'created': created,
            'destroyed': None,
            'destroyed_by_instance_id': None,
            'destroyed_building_percent': None,
            'deleted': False,
            'created_x': pos_x,
            'created_y': pos_y,
            'destroyed_x': None,
            'destroyed_y': None,
            'building_started': None,
            'building_completed': None,
            'total_idle_time': None
        }


def update_objects(tick, owner_id, instance_id, class_id, object_id, killed, deleted, pos_x, pos_y, percent, killed_by, start, complete, idle, tech_id, objects, state, last):
    """Update object/state structures."""
    player_number = owner_id if owner_id and owner_id > 0 else None
    if instance_id not in objects:
        return
    if not class_id:
        class_id = objects[instance_id]['initial_class_id']
    if not killed:
        killed = objects[instance_id].get('killed')
    data = {}
    if class_id == CLASS_UNIT and pos_x is not None and pos_y is not None:
        data.update({
            'destroyed_x': pos_x,
            'destroyed_y': pos_y
        })

    if killed and killed > 0 and instance_id in objects:
        data.update({
            'destroyed': killed,
            'deleted': deleted
        })
        if class_id == CLASS_BUILDING:
            data['destroyed_building_percent'] = percent
        if killed in objects:
            data.update({
                'destroyed_by_instance_id': killed_by
            })
    objects[instance_id].update(data)

    if class_id == CLASS_BUILDING and instance_id in objects:
        if start and start > 0 and objects[instance_id]['building_started'] is None:
            objects[instance_id]['building_started'] = start
        if complete and complete > 0 and objects[instance_id]['building_completed'] is None:
            objects[instance_id]['building_completed'] = complete

    if instance_id in objects and idle:
        objects[instance_id]['total_idle_time'] = int(idle)

    researching_technology_id = tech_id if tech_id and tech_id > 0 else None
    change = (
        instance_id not in last or
        has_diff(
            last[instance_id],
            player_number=player_number,
            class_id=class_id,
            object_id=object_id,
            researching_technology_id=researching_technology_id
        )
    )
    snapshot = {
        'timestamp': tick,
        'instance_id': instance_id,
        'player_number': player_number,
        'object_id': object_id,
        'class_id': class_id,
        'researching_technology_id': researching_technology_id
    }
    if change:
        state.append(snapshot)
    last[instance_id] = snapshot


def enrich_actions(actions, objects, states):
    """Attach player to anonymous actions."""
    owners = defaultdict(list)
    for state in states:
        owners[state['instance_id']].append((state['timestamp'], state['player_number']))
    last_tick = None
    last_seen = []
    for (tick, action_type, payload) in actions:
        if 'player_id' not in payload and 'object_ids' in payload:
            for object_id in payload['object_ids']:
                if object_id not in owners:
                    continue
                for (timestamp, player_number) in owners[object_id]:
                    if timestamp > tick:
                        payload['player_id'] = player_number
                        break
                if object_id not in objects:
                    continue
                if 'player_id' not in payload:
                    payload['player_id'] = objects[object_id]['initial_player_number']
                    break
        action = (tick, action_type, payload)
        if tick != last_tick:
            last_seen = []
            last_tick = tick
        if action not in last_seen and 'player_id' in payload and payload['player_id'] is not None:
            yield action
            last_seen.append(action)


def add_action_count(timeseries, actions):
    i = iter(actions)
    tick, _, payload = next(i)
    last_ts = 0
    for record in timeseries:
        if record['timestamp'] > last_ts:
            count = defaultdict(int)
        while tick < record['timestamp']:
            count[payload['player_id']] += 1
            try:
                tick, _, payload = next(i)
            except StopIteration:
                break
        record['action_count'] = count[record['player_number']]
        last_ts = record['timestamp']

def _dist(a, b):
    return math.sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)

def add_map_control(timeseries, action_list, players, teams, duration):
    #opps[pid] = {pid: xy}
    coords = {}
    lookup = {player['number']:player for player in players}
    opps = defaultdict(list)

    for i, team_a in enumerate(teams):
        for player_number in team_a:
            coords[player_number] = lookup[player_number]['position']
            for j, team_b in enumerate(teams):
                if i == j:
                    continue
                opps[player_number] += team_b
                #for opponent_number in team_b:
                #    opps[player_number] += team_b[opponent_number] = lookup[opponent_number]['position']

    percents = defaultdict(list)

    last = {}
    dk = {}
    for player in players:
        last[player['number']] = [] #deque(maxlen=30)
        dk[player['number']] = deque(maxlen=30)


    def next_action(act):
        tick, _, payload = next(act)
        if 'x' not in payload or 'y' not in payload or 'player_id' not in payload:
            return tick, None, None
        to_me = _dist((payload['x'], payload['y']), coords[payload['player_id']])
        options = []
        for opp in opps[payload['player_id']]:
            to_opp = _dist((payload['x'], payload['y']), coords[opp])
            between = _dist(coords[payload['player_id']], coords[opp])
            options.append(round((((to_me - to_opp) + between) / (between * 2)) * 100))
        return tick, payload['player_id'], max(options)


    it = iter(action_list)
    tick, player_number, control = next_action(it)
    for record in timeseries:
        while tick < record['timestamp']:
            if control:
                if len(last[player_number]) == 30:
                    last[player_number] = []
                last[player_number].append(control)
                mc = last[record['player_number']]
                dk[player_number].append(sum(mc)/len(mc) if len(mc) > 0 else 0)
            try:
                tick, player_number, control = next_action(it)
            except StopIteration:
                break
        mc = dk[record['player_number']]
        record['map_control'] = sum(mc)/len(mc) if len(mc) > 0 else 0
        print(record['timestamp'], record['player_number'], record['map_control'])

def transform_objects(objects):
    """Transform objects."""
    obj_list = []
    for i, obj in objects.items():
        data = dict(obj, instance_id=i)
        if data['destroyed'] is None:
            data['destroyed_x'] = None
            data['destroyed_y'] = None
        obj_list.append(data)
    return obj_list


def transform_seed_objects(objects):
    """Map seed objects to state format."""
    return {obj['instance_id']: {
        'initial_player_number': obj['player_number'],
        'initial_object_id': obj['object_id'],
        'initial_class_id': obj['class_id'],
        'created': 0,
        'created_x': obj['x'],
        'created_y':obj['y'],
        'destroyed': None,
        'destroyed_by_instance_id': None,
        'destroyed_building_percent': None,
        'deleted': False,
        'destroyed_x': None,
        'destroyed_y': None,
        'building_started': None,
        'building_completed': None,
        'total_idle_time': None
    } for obj in objects}


async def get_extracted_data(start_time, duration, playback, handle, interval, seed_objects, players, teams):
    """Get extracted data."""
    timeseries = []
    research = defaultdict(dict)
    market = []
    objects = transform_seed_objects(seed_objects)
    state = []
    last = {}
    actions = []
    version = None
    start = time.time()
    client = await Client.create(playback, handle.name, start_time, duration, interval)
    async for tick, source, message in client.sync():
        if source == Source.MEMORY:
            if not version:
                version = message.StateReaderVersion().decode('ascii')

            update_market(tick, message.MarketCoefficients().Food(), message.MarketCoefficients().Wood(), message.MarketCoefficients().Stone(), market)

            # Add any new objects before updating to ensure fks are present for updates
            for i in range(0, message.ObjectsLength()):
                obj = message.Objects(i)
                add_objects(obj.OwnerId(), obj.Id(), obj.MasterObjectClass(), obj.MasterObjectId(), obj.CreatedTime(), obj.Position().X(), obj.Position().Y(), objects)

            for i in range(0, message.ObjectsLength()):
                obj = message.Objects(i)
                update_objects(
                    tick, obj.OwnerId(), obj.Id(), obj.MasterObjectClass(), obj.MasterObjectId(), obj.KilledTime(),
                    obj.DeletedByOwner(), obj.Position().X(), obj.Position().Y(), obj.BuildingPercentComplete(),
                    obj.KilledByUnitId(), obj.BuildingStartTime(), obj.BuildingCompleteTime(), obj.CumulativeIdleTime(),
                    obj.CurrentlyResearchingTechId(), objects, state, last
                )

            for i in range(0, message.PlayersLength()):
                player = message.Players(i)
                timeseries.append(build_fb_timeseries_record(tick, player))
                for j in range(0, player.TechsLength()):
                    tech = player.Techs(j)
                    update_research(
                        player.Id(), tech.State(), tech.IdIndex(), tech.Time(), research,
                        TECH_STATE_RESEARCHING, TECH_STATE_DONE, TECH_STATE_AVAILABLE
                    )

        elif source == Source.MGZ:
            if message[0] == fast.Operation.ACTION:
                actions.append((tick, *message[1]))

    handle.close()

    action_list = list(enrich_actions(actions, objects, state))
    add_action_count(timeseries, action_list)
    add_map_control(timeseries, action_list, players, teams, duration)

    return {
        'version': version,
        'runtime': timedelta(seconds=int(time.time() - start)),
        'timeseries': timeseries,
        'research': flatten_research(research),
        'market': market,
        'objects': transform_objects(objects),
        'state': state,
        'actions': action_list,
        'winners': set()
    }


def external_extracted_data(data, seed_objects, players, teams, actions):
    """Merge externally-sourced extracted data."""
    research = defaultdict(dict)
    objects = transform_seed_objects(seed_objects)
    state = []
    market = []
    last = {}
    timeseries = []
    winners = set()
    version = data['version']
    interval = data['messageInterval']
    for message in data['messages']:
        update_market(message['time'], message['world']['foodPrice'], message['world']['woodPrice'], message['world']['stonePrice'], market)

        # Add any new objects before updating to ensure fks are present for updates
        for obj in message['objects']:
            # Only add objects the first time we see them
            if 'ownerId' not in obj or 'masterObjectClass' not in obj:
                continue
            add_objects(
                obj['ownerId'], obj['id'], obj['masterObjectClass'], obj['masterObjectId'], obj['createdTime'],
                obj['position']['x'], obj['position']['y'], objects
            )

        for obj in message['objects']:
            update_objects(
                message['time'], obj.get('ownerId'), obj['id'], obj.get('masterObjectClass'), obj.get('masterObjectId'),
                obj.get('killedTime'), None, obj.get('position', {}).get('x'), obj.get('position', {}).get('y'), obj.get('buildingPercentComplete'),
                None, obj.get('buildingStartTime'), obj.get('buildingCompleteTime'), obj.get('cumulativeIdleTime'), None, objects, state, last
            )

        for player in message['players'][1:]:
            timeseries.append(build_json_timeseries_record(message['time'], player))
            if player['status'] == STATUS_WINNER:
                winners.add(player['id'])

        for event in message['events']:
            if event['data']['tag'] == 'techStateChange':
                update_research(
                    event['data']['playerId'], event['data']['state'], event['data']['index'], event['worldTime'],
                    research, DE_TECH_STATE_RESEARCHING, DE_TECH_STATE_DONE, DE_TECH_STATE_AVAILABLE
                )

    for obj in objects.values():
        if not obj['destroyed']:
            obj['destroyed_x'] = None
            obj['destroyed_y'] = None

    action_list = list(enrich_actions(actions, objects, state))
    add_action_count(timeseries, action_list)

    return {
        'version': version,
        'interval': interval,
        'runtime': None,
        'timeseries': timeseries,
        'research': flatten_research(research),
        'market': market,
        'objects': [dict(obj, instance_id=i) for i, obj in objects.items()],
        'state': state,
        'actions': action_list,
        'winners': winners
    }
